import json
from collections.abc import Mapping

from pyarxaas.models.request_builder import RequestBuilder
from pyarxaas.arxaas_connector import ARXaaSConnector
from pyarxaas.models.anonymize_result import AnonymizeResult
from pyarxaas.models.dataset import Dataset
from pyarxaas.models.risk_profile import RiskProfile


class ARXaaS:
    """
    Represents the connection to ARXaaS. All public methods result in a call to the service.
    """

    def __init__(self, url: str, connector=ARXaaSConnector, client=None):
        self._connector = connector(url, client=client)
        self._connector.test_connection()

    def anonymize(self, dataset: Dataset, privacy_models,suppression_limit: float = None) -> AnonymizeResult:
        """
        Attempt to anonymize a dataset with provided privacy models

        :param dataset: Dataset to be anonymized
        :param privacy_models: privacy models to be used in the anonymization
        :param suppression_limit: suppression limit to be used in the anonymization
        :return: Dataset with anonymized data
        """
        request_payload = self._anonymize_payload(dataset, privacy_models, suppression_limit)
        response = self._anonymize(request_payload)
        return self._anonymize_result(response)

    def _anonymize_payload(self, dataset, privacy_models, suppression_limit) -> Mapping:
        """
        Creates a anonymize payload to be sent to the backend

        :param dataset: Dataset to be anonymized
        :param privacy_models: privacy models to be used in the anonymization
        :param suppression_limit: suppression limit to be used in the anonymization
        :return: Mapping payload
        """

        return RequestBuilder(dataset)\
            .add_privacy_models(privacy_models)\
            .add_suppression_limit(suppression_limit)\
            .build_anonymize_request()

    def _anonymize(self, payload):
        """
        Passes the payload to the connector for anonymization
        :param payload: Mapping matching the service anonymize json schema
        :return:
        """
        response = self._connector.anonymize_data(payload)
        return response

    def _anonymize_result(self, response):
        """
        Creates the result to be delivered back to the caller

        :param response:
        :return:
        """
        json_string = response.text
        response_dict = json.loads(json_string)
        attributes = self._attributes(response_dict)
        dataset = Dataset(response_dict["anonymizeResult"]["data"], attributes)
        risk_profile = RiskProfile(response_dict["riskProfile"])
        anon_status = response_dict["anonymizeResult"]["anonymizationStatus"]
        anonymization_metrics = response_dict["anonymizeResult"]["metrics"]
        return AnonymizeResult._from_response(dataset, risk_profile, anonymization_metrics, anon_status)

    def risk_profile(self, dataset: Dataset) -> RiskProfile:
        """
        Creates a risk profile for a provided Dataset

        RiskProfile contains:
         - re-identifiaction risks
         - distributed risk

        :param dataset: Dataset to create a risk profile for
        :return: RiskProfile
        """

        analyze_request = self._risk_profile_payload(dataset)
        response = self._risk_profile(analyze_request)
        metric_dict = json.loads(response.text)
        return RiskProfile(metric_dict)

    def _risk_profile_payload(self, dataset):
        """
        Creates a risk profile payload to be sent to the backend

        :param dataset: Dataset to be analyzed
        :return: Mapping payload
        """

        return RequestBuilder(dataset).build_analyze_request()

    def _risk_profile(self, data_dict):
        """
        Passes the payload to the connector for risk profiling
        :param payload: Mapping matching the service anonlyze json schema
        :return: Response
        """

        response = self._connector.risk_profile(data_dict)
        return response

    def _attributes(self, response_dict):
        raw = response_dict["anonymizeResult"]["attributes"]
        attribute_dict = {attribute["field"]: attribute["attributeTypeModel"] for attribute in raw}
        return attribute_dict

    def hierarchy(self, redaction_builder, column):
        """
        Creates a value generalization hierarchy with the passed in builder for the passed in column.


        :param redaction_builder: a Hierarchy builder instance
        :param column: a list of values
        :return: list[list] containing the created hierarchy
        """

        request = redaction_builder._request_payload()
        request["column"] = column
        response = self._connector.hierarchy(request)
        response_dict = json.loads(response.text)
        return response_dict["hierarchy"]


from pyarxaas.privacy_models import LDiversityDistinct, LDiversityGrassbergerEntropy, LDiversityShannonEntropy, LDiversityRecursive
import unittest


class PrivacyModelsLDiversityTest(unittest.TestCase):

    def test_privacyModelIterator(self):
        self.privacymodel = LDiversityDistinct(l=3, column_name="column")
        self.assertEqual(list(iter(self.privacymodel))[1], "column_name", "Iterator does not create correct list of values")


    def test_lenght(self):
        self.privacymodel = LDiversityDistinct(l=3, column_name="column")
        self.assertEqual(len(self.privacymodel), 2)


    def test_message(self):
        self.privacymodel = LDiversityDistinct(l=3, column_name="column")
        self.assertEqual(str(self.privacymodel), "LDiversityDistinct(l=3, column_name=column)")

    def test_set_data_distinct(self):
        self.distict = LDiversityDistinct(l=3, column_name="sensitive_column_name")
        self.assertEqual(self.distict._anonymity_name, "LDIVERSITY_DISTINCT", "Recursive model name not correct")

    def test_set_data_grassberger(self):
        self.grassberger = LDiversityGrassbergerEntropy(l=3, column_name="sensitive_column_name")
        self.assertEqual(self.grassberger._anonymity_name, "LDIVERSITY_GRASSBERGERENTROPY", "Grassberger model name not correct")

    def test_set_data_shannonentropy(self):
        self.shannon_entropy = LDiversityShannonEntropy(l=3, column_name="sensitive_column_name")
        self.assertEqual(self.shannon_entropy.get("l"), 3, "l_value does not equals 3")
        self.assertNotEqual(self.shannon_entropy.get("l"), 1, "l_Value does equals 1")
        self.assertEqual(self.shannon_entropy.get("column_name"), "sensitive_column_name", "Columnname is set correctly")
        self.assertNotEqual(self.shannon_entropy.get("column_name"), "unsensitive_column_name", "Columnname is not set incorrectly")
        self.assertEqual(self.shannon_entropy._anonymity_name, "LDIVERSITY_SHANNONENTROPY", "Anonymity name not set correctly")

    def test_set_data_recursive(self):
        self.recursive = LDiversityRecursive(l=3, c=1, column_name="sensitive_column_name")
        self.assertEqual(self.recursive._anonymity_name, "LDIVERSITY_RECURSIVE", "Recursive model name not correct")



from uplink import Consumer, get, headers, Path, Query, post, Body, json
import logging

from uplink import response_handler
from requests.exceptions import RequestException, HTTPError


@response_handler
def raise_for_status(response):
    """
    Raises exceptions for not successful response status codes
    :param response: response from service
    :return:
    """
    if 400 <= response.status_code < 500:
        raise RequestException(response.text)
    if response.status_code >= 500:
        raise HTTPError(response.text)
    return response


class ARXaaSConnector(Consumer):
    """ Understands connection to ARXaaS endpoints"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._logger = logging.getLogger(__name__)

    @raise_for_status
    @json
    @post("api/anonymize")
    def anonymize_data(self, payload: Body):
        """Post data to AaaS Backend"""

    @raise_for_status
    @json
    @post("api/analyze")
    def risk_profile(self, payload: Body):
        """Post data to AaaS Backend"""

    @raise_for_status
    @json
    @post("api/hierarchy")
    def hierarchy(self, payload: Body):
        """Post data to AaaS Backend"""

    @raise_for_status
    @get("/")
    def root(self):
        """ Get root of service """

    def test_connection(self):
        url = self.session.base_url
        logger = self._logger
        try:
            response = self.root()
            logger.info(f"Connected to url={url} status={response.status_code}")
        except Exception as err:
            logger.warning(f"Cannot connect to url={url} message={err}")




from pyarxaas.privacy_models import TClosenessEqualDistance, TClosenessOrderedDistance
import unittest


class PrivacyModelsTclosenessTest(unittest.TestCase):

    def test_privacyModelIterator(self):
        self.privacymodel = TClosenessEqualDistance(t=3, column_name="column")
        self.assertEqual(list(iter(self.privacymodel))[1], "column_name", "Iterator does not create correct list of values")

    def test_lenght(self):
        self.privacymodel = TClosenessEqualDistance(t=3, column_name="column")
        self.assertEqual(len(self.privacymodel), 2)

    def test_message(self):
        self.privacymodel = TClosenessEqualDistance(t=3, column_name="column")
        self.assertEqual(str(self.privacymodel), "TClosenessEqualDistance(t=3, column_name=column)")

    def test_set_data_EqualDistance(self):
        self.distict = TClosenessEqualDistance(t=3, column_name ="sensitive_column_name")
        self.assertEqual(self.distict._anonymity_name, "TCLOSENESS_EQUAL_DISTANCE", "Recursive model name not correct")

    def test_set_data_TClosenessOrderedDistance(self):
        self.distict = TClosenessOrderedDistance(t=3, column_name="sensitive_column_name")
        self.assertEqual(self.distict._anonymity_name, "TCLOSENESS_ORDERED_DISTANCE", "Recursive model name not correct")


from collections.abc import Mapping, Callable
from IPython.core import display as ICD
import pandas


def print_privacy_models(models: Mapping, printer: Callable) -> pandas.DataFrame:
    """
    Prints out a human readable view of the provided Privacy Models

    :param models: collection of Privacy models
    :param printer: callable
    :return: None
    """
    for model in models.values():
        return _model_to_dataframe(model)


def print_mapping(name_data_mapping: Mapping):
    """
    Prints the current content of the payload object to stdout

    :param name_data_mapping: Mapping object name:data
    :return: None
    """

    for name, dataframe in name_data_mapping.items():
        print(name)
        print(dataframe)
        print("-"*40 + "\n")


def jupyter_print_mapping(name_dataframe_mapping: Mapping):
    """
    Print the provided name_dataframe_mapping in a J

    :param name_dataframe_mapping:
    :return:
    """
    for name, dataframe in name_dataframe_mapping.items():
        print(name)
        ICD.display(dataframe)
        print("-"*40)


def _model_to_dataframe(model):
    return pandas.DataFrame(model.items())



import unittest

from pyarxaas.models.anonymization_metrics import AnonymizationMetrics
from pyarxaas.models.anonymize_result import AnonymizeResult
from pyarxaas.models.dataset import Dataset
from tests.pyarxaas import data_generator


class AnonymizeResultTest(unittest.TestCase):

    def setUp(self):
        self.test_dataset = data_generator.id_name_dataset()
        self.test_risk_profile = data_generator.risk_profile()
        self.raw_metrics = data_generator.raw_anonymization_metrics()
        self.test_anon_metrics = AnonymizationMetrics(self.raw_metrics)

    def test_init(self):
        AnonymizeResult(self.test_dataset, self.test_risk_profile, self.test_anon_metrics, 'ANONYMOUS')

    def test_equaltiy(self):
        ar1 = AnonymizeResult(self.test_dataset, self.test_risk_profile, self.test_anon_metrics, 'ANONYMOUS')
        ar2 = AnonymizeResult(self.test_dataset, self.test_risk_profile, self.test_anon_metrics, 'ANONYMOUS')
        self.assertEqual(ar1, ar2)
        ar2._dataset = Dataset([["data", "data2"]])
        self.assertNotEqual(ar1, ar2)

    def test_hash(self):
        ar1 = AnonymizeResult(self.test_dataset, self.test_risk_profile, self.test_anon_metrics, 'ANONYMOUS')
        ar2 = AnonymizeResult(self.test_dataset, self.test_risk_profile, self.test_anon_metrics, 'ANONYMOUS')
        test_set = {ar1, ar2}
        self.assertEqual(1, len(test_set))

    def test_dataset(self):
        ar = AnonymizeResult(self.test_dataset, self.test_risk_profile, self.test_anon_metrics, 'ANONYMOUS')
        dataset = ar.dataset
        self.assertIsNotNone(dataset)
        self.assertEqual(self.test_dataset, dataset)

    def test_risk_profile(self):
        ar = AnonymizeResult(self.test_dataset, self.test_risk_profile, self.test_anon_metrics, 'ANONYMOUS')
        risk_profile = ar.risk_profile
        self.assertIsNotNone(risk_profile)
        self.assertEqual(self.test_risk_profile, risk_profile)

    def test_anonymization_metrics(self):
        ar = AnonymizeResult(self.test_dataset, self.test_risk_profile, self.test_anon_metrics, 'ANONYMOUS')
        anonymization_metrics = ar.anonymization_metrics
        self.assertIsNotNone(anonymization_metrics)
        self.assertEqual(self.test_anon_metrics, anonymization_metrics)

    def test_create_from_response_payload(self):
        ar1 = AnonymizeResult._from_response(self.test_dataset, self.test_risk_profile, self.raw_metrics, 'ANONYMOUS')
        ar2 = AnonymizeResult(self.test_dataset, self.test_risk_profile, self.test_anon_metrics, 'ANONYMOUS')
        self.assertEqual(ar1, ar2)




from collections.abc import Mapping, Sequence

from pyarxaas.privacy_models import PrivacyModel
from pyarxaas.models.dataset import Dataset


class RequestBuilder:
    """
    Builds JSON serializable mappings from Dataset and PrivacyModels
    """

    def __init__(self, dataset: Dataset):
        self._dataset = dataset
        self._privacy_models = []
        self._suppression_limit = None

    def add_privacy_model(self, privacy_model: PrivacyModel):
        self._privacy_models.append(privacy_model)
        return self

    def add_privacy_models(self, privacy_models: Sequence):
        self._privacy_models += privacy_models
        return self

    def add_suppression_limit(self, suppression_limit):
        self._suppression_limit = suppression_limit
        return self

    def build_analyze_request(self):
        data_dict = self._dataset._payload()
        self._strip_hierarchies(data_dict)
        return data_dict

    def build_anonymize_request(self) -> Mapping:
        """
        Creates a anonymize payload to be sent to the backend

        :param dataset: Dataset to be anonymized
        :param privacy_models: privacy models to be used in the anonymization
        :param suppression_limit: suppression limit to be used in the anonymization
        :return: Mapping payload
        """
        data_dict = self._dataset._payload()
        models = []
        for model in self._privacy_models:
            models.append(model._payload())
        data_dict["privacyModels"] = models
        data_dict["suppressionLimit"] = self._suppression_limit
        return data_dict

    def _strip_hierarchies(self, data_dict) -> None:
        attributes = data_dict["attributes"]
        for attribute in attributes:
            del attribute["hierarchy"]






import pandas


class Data:
    """
    Represents a collection of dataset rows
    """

    def __init__(self, headers, rows):
        self._headers = headers
        self._rows = rows

    @property
    def payload(self):
        return [self._headers] + self._rows

    @property
    def dataframe(self):
        return pandas.DataFrame(self._rows, columns=self._headers)

    def describe(self, indent):
        print("data:")
        print(self._describe_data_headers(indent))
        print("rows:")
        print(self._describe_data_rows(indent))

    def _describe_data_headers(self, indent):
        string = " "*indent + "headers:\n"
        string += " " * indent*2 + str(self._headers)
        return string

    def _describe_data_rows(self, indent):
        indent = indent*2
        max_rows_to_print = 5
        rows_to_print = min(max_rows_to_print, (len(self._rows)))
        max_columns_to_print = 8
        string = ""
        for row in self._rows[0: rows_to_print]:
            if len(row) > max_columns_to_print:
                columns_to_print_mid = max_columns_to_print // 2
                print_row = row[:columns_to_print_mid]
                print_row.append("...")
                print_row += row[-columns_to_print_mid:]
                row = print_row
            string += " "*indent +\
                      str(row) +\
                      "\n"
        if len(self._rows) > max_rows_to_print:
            string += " "*indent + str("...")
        return string

    def __hash__(self):
        d_hash = hash("")
        for header in self._headers:
            d_hash = hash(d_hash + hash(header))
        for row in self._rows:
            for cell in row:
                d_hash = hash(d_hash + hash(cell))
        return d_hash



import copy

from pyarxaas.hierarchy.grouping_based_hierarchy import GroupingBasedHierarchy
from pyarxaas.hierarchy.interval_builder.interval import Interval


class IntervalHierarchyBuilder(GroupingBasedHierarchy):
    """
    Represents a specific strategy for creating a value generalization hierarchy
    """

    def __init__(self):
        super().__init__()
        self._intervals = {}

    @property
    def intervals(self):
        return list(copy.deepcopy(self._intervals))

    def add_interval(self, from_n, to_n, label: str=None):
        """
        Add a interval to the builder. from_n is inclusive, to_n is exclusive

        :param from_n: create interval inclusive from this value
        :param to_n: create interval to this value
        :param label: (optional) set a string label for the interval
        :return: None

        """
        # using dict to enforce uniqueness and order
        self._intervals[Interval(from_n, to_n, label)] = ""

    def _request_payload(self):
        return {
            "builder": {
                "type": "intervalBased",
                "dataType": self._data_type(),
                "intervals": [interval.payload() for interval in self._intervals.keys()],
                "levels": [level.payload() for level in self.levels]
            }
        }

    def _data_type(self):
        for interval in self._intervals:
            if interval.is_decimal():
                return "DOUBLE"
        return "LONG"

    @staticmethod
    def _is_decimal(number):
        return "." in number



import unittest

from pyarxaas.privacy_models import KAnonymity
from pyarxaas.models.request_builder import RequestBuilder
from tests.pyarxaas import data_generator


class RequestBuilderTest(unittest.TestCase):

    def setUp(self):
        self.test_dataset = data_generator.id_name_dataset()
        self.kanon = KAnonymity(4)
        self.suplimit = 0.001
        self.expected_analyze_request = {'data': [['id', 'name'], ['0', 'Viktor'], ['1', 'Jerry']],
                                         'attributes': [{'field': 'id', 'attributeTypeModel': 'IDENTIFYING'},
                                                        {'field': 'name', 'attributeTypeModel': 'QUASIIDENTIFYING'}]}
        self.expected_anonymize_request = {'data': [['id', 'name'], ['0', 'Viktor'], ['1', 'Jerry']],
                                           'attributes': [{'field': 'id', 'attributeTypeModel': 'IDENTIFYING', 'hierarchy': None},
                                                          {'field': 'name', 'attributeTypeModel': 'QUASIIDENTIFYING', 'hierarchy': None}],
                                           'privacyModels': [{'privacyModel': 'KANONYMITY', 'params': {'k': 4}}],
                                           'suppressionLimit': 0.001}

    def test_init(self):
        RequestBuilder(self.test_dataset)

    def test_add_privacy_model(self):
        req_builder = RequestBuilder(self.test_dataset)
        req_builder.add_privacy_model(self.kanon)

    def test_add_privacy_model(self):
        req_builder = RequestBuilder(self.test_dataset)
        req_builder.add_suppression_limit(self.suplimit)

    def test_build_analyze_request(self):
        analyze_request = RequestBuilder(self.test_dataset).build_analyze_request()
        self.assertEqual(self.expected_analyze_request, analyze_request)

    def test_build_anonymize_request(self):
        anonymize_request = RequestBuilder(self.test_dataset)\
            .add_privacy_model(self.kanon)\
            .add_suppression_limit(self.suplimit)\
            .build_anonymize_request()
        self.assertEqual(self.expected_anonymize_request, anonymize_request)


# module for collection of test data generating functions and classes
import json
from os import path
from pyarxaas.models.anonymize_result import AnonymizeResult
from pyarxaas.models.attribute_type import AttributeType
from pyarxaas.models.dataset import Dataset
from pyarxaas.models.risk_profile import RiskProfile

here = path.abspath(path.dirname(__file__))


def response_dict(file):
    with open(path.join(here, file), encoding="utf-8") as json_file:
        return json.load(json_file)


def analyze_response():
    return response_dict("test_data/analyze_response_test_data.json")


def anonymize_response():
    return response_dict("test_data/anonymize_response_test_data.json")


def id_name_dataset() -> Dataset:
    test_data = [['id', 'name'],
                      ['0', 'Viktor'],
                      ['1', 'Jerry']]
    test_attribute_type_mapping = {'id': AttributeType.IDENTIFYING,
                                        'name': AttributeType.QUASIIDENTIFYING}
    return Dataset(test_data, test_attribute_type_mapping)


def raw_anonymization_metrics():
    metrics = anonymize_response()["anonymizeResult"]["metrics"]
    return metrics


def risk_profile() -> RiskProfile:
    raw_risk_profile = analyze_response()
    return RiskProfile(raw_risk_profile)



import unittest

from pyarxaas.hierarchy import IntervalHierarchyBuilder


class IntervalHierarchyBuilderTest(unittest.TestCase):

    def test_init(self):
        IntervalHierarchyBuilder()

    def test_add_level(self):
        ib = IntervalHierarchyBuilder()
        ib.level(0).add_group(2, "test_label")
        ib.level(1).add_group(4)

    def test_add_levels_not_availible_raises_exception(self):
        with self.assertRaises(AttributeError):
            ib = IntervalHierarchyBuilder()
            ib.level(1).add_group(2)

        with self.assertRaises(AttributeError):
            ib = IntervalHierarchyBuilder()
            ib.level(0).add_group(2)
            ib.level(2).add_group(1)

    def test__request_payload(self):
        expected = {
            "builder" : {
                "type" : "intervalBased",
                "dataType": "LONG",
                "intervals": [],
                "levels" : [ {
                    "level" : 0,
                    "groups" : [ {
                        "grouping" : 2,
                        "label" : "test_label"
                    } ]
                } ]
            }
        }

        ib = IntervalHierarchyBuilder()
        ib.level(0).add_group(2, "test_label")
        payload = ib._request_payload()
        self.assertEqual(expected, payload)

    def test_intervals_are_unique(self):
        ib = IntervalHierarchyBuilder()
        ib.add_interval(0, 10)
        ib.add_interval(0, 10)
        self.assertEqual(1, len(ib.intervals))

    def test_intervals_order(self):
        ib = IntervalHierarchyBuilder()
        ib.add_interval(0, 10)
        ib.add_interval(10, 20)
        ib.add_interval(20, 30)
        self.assertEqual(10, ib.intervals[0]._to)
        self.assertEqual(20, ib.intervals[1]._to)
        self.assertEqual(30, ib.intervals[2]._to)

    def test_group_building(self):
        ib = IntervalHierarchyBuilder()
        ib.level(0)\
            .add_group(2)\
            .add_group(2)
        ib.level(1)\
            .add_group(2)
        self.assertEqual(2, len(ib.levels[0]._groups))
        self.assertEqual(1, len(ib.levels[1]._groups))

    def test_decimal_interval(self):
        expected = {'builder':
                        {'type': 'intervalBased',
                         'dataType': 'DOUBLE',
                         'intervals':
                             [{'from': 0.0, 'to': 10, 'label': None},
                              {'from': 10, 'to': 20, 'label': None},
                              {'from': 20, 'to': 30, 'label': None}],
                         'levels': []}}

        ib = IntervalHierarchyBuilder()
        ib.add_interval(0.0, 10)
        ib.add_interval(10, 20)
        ib.add_interval(20, 30)
        payload = ib._request_payload()
        self.assertEqual(expected, payload)

    def test_integer_interval(self):
        expected = {'builder':
                        {'type': 'intervalBased',
                         'dataType': 'LONG',
                         'intervals':
                             [{'from': 0.0, 'to': 10, 'label': None},
                              {'from': 10, 'to': 20, 'label': None},
                              {'from': 20, 'to': 30, 'label': None}],
                         'levels': []}}

        ib = IntervalHierarchyBuilder()
        ib.add_interval(0, 10)
        ib.add_interval(10, 20)
        ib.add_interval(20, 30)
        payload = ib._request_payload()
        self.assertEqual(expected, payload)






import json
import unittest

from uplink import Body

from pyarxaas.hierarchy import RedactionHierarchyBuilder
from pyarxaas.arxaas_connector import ARXaaSConnector
from pyarxaas.privacy_models import KAnonymity
from pyarxaas.arxaas import ARXaaS
from pyarxaas.models.anonymize_result import AnonymizeResult
from pyarxaas.models.attribute_type import AttributeType
from pyarxaas.models.dataset import Dataset
from tests.pyarxaas import data_generator


class AnalyzationResponseStub:

    @property
    def text(self):
        return json.dumps(data_generator.analyze_response())

    @property
    def status_code(self):
        return 200


class AnonymzationResponseStub:

    @property
    def text(self):
        return json.dumps(data_generator.anonymize_response())

    @property
    def status_code(self):
        return 200


class HierarchyResponseStub:

    @property
    def text(self):
        return '{"hierarchy":[["1123","*123","**23","***3","****"],["1321","*321","**21","***1","****"],["1234",' \
               '"*234","**34","***4","****"],["1532","*532","**32","***2","****"]]} '

    @property
    def status_code(self):
        return 200


class RootResponseStub:

    @property
    def text(self):
        return '{"_links": {_self: "test"}}'

    @property
    def status_code(self):
        return 200


class MockAaasConnector(ARXaaSConnector):

    def anonymize_data(self, payload: Body):
        return AnonymzationResponseStub()

    def risk_profile(self, payload: Body):
        return AnalyzationResponseStub()

    def hierarchy(self, payload: Body):
        return HierarchyResponseStub()

    def root(self):
        return RootResponseStub()


class ARXaaSTest(unittest.TestCase):

    def setUp(self):
        self.test_data = [['id', 'name'],
                          ['0', 'Viktor'],
                          ['1', 'Jerry']]
        self.test_attribute_type_mapping = {'id': AttributeType.IDENTIFYING,
                                            'name': AttributeType.QUASIIDENTIFYING}

        self.test_dataset = Dataset(self.test_data, self.test_attribute_type_mapping)

        self.test_raw_analyze_response = data_generator.analyze_response()
        self.test_raw_anon_response = data_generator.anonymize_response()

    def test_init(self):
        ARXaaS('http://localhost', connector=MockAaasConnector)

    def test_analyze(self):
        aaas = ARXaaS('http://localhost', connector=MockAaasConnector)
        self.assertIsNotNone(aaas.risk_profile(self.test_dataset))

    def test_anaonymize(self):
        aaas = ARXaaS('http://localhost', connector=MockAaasConnector)
        self.assertIsNotNone(aaas.anonymize(self.test_dataset, privacy_models=[KAnonymity(4)]))

    def test_risk_profile_return_value(self):
        aaas = ARXaaS('http://localhost', connector=MockAaasConnector)
        risk_profile = aaas.risk_profile(self.test_dataset)
        df = risk_profile.re_identification_risk_dataframe()
        self.assertEqual(self.test_raw_analyze_response["reIdentificationRisk"]["measures"][
                             "records_affected_by_highest_prosecutor_risk"],
                         df["records_affected_by_highest_prosecutor_risk"][0])

    def test_anonymize_return_value(self):
        aaas = ARXaaS('http://localhost', connector=MockAaasConnector)
        anonymized_dataset = aaas.anonymize(self.test_dataset, [KAnonymity(4)])
        self.assertIsInstance(anonymized_dataset, AnonymizeResult)

    def test_anonymize__dataset_attributes_are_correct(self):
        aaas = ARXaaS('http://localhost', connector=MockAaasConnector)
        anonymize_result = aaas.anonymize(self.test_dataset, [KAnonymity(4)])
        self.assertEqual(AttributeType.IDENTIFYING, AttributeType(anonymize_result.dataset._attributes[0].type))

    def test_redaction_based_hierarchy(self):
        expected = [["1123", "*123", "**23", "***3", "****"], ["1321", "*321", "**21", "***1", "****"],
                    ["1234", "*234", "**34", "***4", "****"], ["1532", "*532", "**32", "***2", "****"]]
        aaas = ARXaaS('http://localhost:8080', connector=MockAaasConnector)
        redaction_builder = RedactionHierarchyBuilder(
            " ",
            "*",
            RedactionHierarchyBuilder.Order.RIGHT_TO_LEFT,
            RedactionHierarchyBuilder.Order.LEFT_TO_RIGHT)
        column = ["1123", "1321", "1234", "1532"]
        hierarchy = aaas.hierarchy(redaction_builder, column)
        self.assertEqual(expected, hierarchy)



from collections import Sequence

from pyarxaas.models.attribute_type import AttributeType


class Attribute:
    """
    Understands Dataset field
    """

    def __init__(self, field_name, type):
        self._field_name = field_name
        self._type = type
        self._hierarchy = None

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(hash(self._field_name) + hash(self._hierarchy) + hash(self._type))

    def __str__(self) -> str:
        return f"field_name={self.name}, type={self.type.name}, hierarchy={self.hierarchy}"

    def __repr__(self):
        return f"Attribute(field_name={self.name}, type={self.type}, hierarchy={self.hierarchy})"

    @property
    def name(self):
        return self._field_name

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, attribute_type):
        attribute_type = AttributeType(attribute_type.value)
        self._type = attribute_type

    @property
    def hierarchy(self):
        return self._hierarchy

    @hierarchy.setter
    def hierarchy(self, hierarchy):
        if not isinstance(hierarchy, Sequence):
            raise ValueError(f"hierarchy has to be of type {Sequence}")
        if not self._is_hierarchy_settable():
            raise ValueError(
                f"{self.__class__}(name={self.name}, type={self.type}) has to be of type {AttributeType.QUASIIDENTIFYING}")
        self._hierarchy = hierarchy

    def _is_hierarchy_settable(self):
        return AttributeType.QUASIIDENTIFYING.value == self.type.value

    @property
    def payload(self):
        return {"field": self._field_name, "attributeTypeModel": self._type.value, "hierarchy": self._hierarchy}


import unittest

from pyarxaas.hierarchy import OrderHierarchyBuilder


class OrderHierarchyBuilderTest(unittest.TestCase):

    def test_init(self):
        OrderHierarchyBuilder()

    def test_add_groups(self):
        ob = OrderHierarchyBuilder()
        ob.level(0).add_group(2, "upper-body-clothes")

    def test__request_payload(self):
        expected = {'builder': {'type': 'orderBased', 'levels': [{'level': 0, 'groups': [{'grouping': 2, 'label': 'upper-body-clothes'}]}]}}
        ob = OrderHierarchyBuilder()
        ob.level(0).add_group(2, "upper-body-clothes")
        self.assertEqual(expected, ob._request_payload())

    def test_adding_same_group_twice_is_not_possible(self):
        ob = OrderHierarchyBuilder()
        ob.level(0).add_group(2, "lower-body-clothes")
        ob.level(0).add_group(2, "upper-body-clothes")
        ob.level(0).add_group(2, "lower-body-clothes")
        self.assertEqual(2, len(ob.level(0)._groups))



import unittest

from pyarxaas.hierarchy import RedactionHierarchyBuilder


class RedactionHierarchyBuildTest(unittest.TestCase):

    def test_init(self):
        RedactionHierarchyBuilder(
            " ",
            "*",
            RedactionHierarchyBuilder.Order.RIGHT_TO_LEFT,
            RedactionHierarchyBuilder.Order.LEFT_TO_RIGHT)

    def test__request_payload(self):
        expected = {
                    'builder': {'type': 'redactionBased',
                                'paddingCharacter': ' ',
                                'redactionCharacter': '*',
                                'paddingOrder': 'RIGHT_TO_LEFT',
                                'redactionOrder': 'LEFT_TO_RIGHT'}
                    }
        redaction_builder = RedactionHierarchyBuilder(
            " ",
            "*",
            RedactionHierarchyBuilder.Order.RIGHT_TO_LEFT,
            RedactionHierarchyBuilder.Order.LEFT_TO_RIGHT)
        request_payload = redaction_builder._request_payload()
        self.assertEqual(expected, request_payload)

    def test_redaction_builder_with_zero_params(self):
        expected = {
                    'builder': {'type': 'redactionBased',
                                'paddingCharacter': ' ',
                                'redactionCharacter': '*',
                                'paddingOrder': 'RIGHT_TO_LEFT',
                                'redactionOrder': 'RIGHT_TO_LEFT'}
                    }

        redaction_builder = RedactionHierarchyBuilder()

        request_payload = redaction_builder._request_payload()
        self.assertEqual(expected, request_payload)

    def test_padding_characters_can_only_be_single_char(self):
        with self.assertRaises(AttributeError):
            RedactionHierarchyBuilder(padding_char="ab")

        with self.assertRaises(AttributeError):
            RedactionHierarchyBuilder(redaction_char="ab")




from collections.abc import Mapping
from abc import ABC


class PrivacyModel(ABC, Mapping):
    """
    Documentation of the privacy models implemented in the ARXaaS service and the definition of the parameters
    each privacy model takes.
    """
    def __init__(self):
        self._internal_dict = {}

    def __getitem__(self, item):
        return self._internal_dict[item]

    def __len__(self) -> int:
        return len(self._internal_dict)

    def __iter__(self):
        return iter(self._internal_dict)

    @property
    def name(self) -> str:
        return self._anonymity_name

    def __str__(self):
        return self._print_message

    def _payload(self):
        return {"privacyModel": self.name, "params": self._internal_dict}


class KAnonymity(PrivacyModel):
    """
    Configuration class for K-Anonymity

    :param k: Value of K  to anonymize the dataset. K must have a value of 2 or higher to take effect.

    """

    def __init__(self, k):
        self._internal_dict = {"k": k}
        self._anonymity_name = "KANONYMITY"
        self._print_message = f"KAnonymity(k={k})"


class LDiversityDistinct(PrivacyModel):
    """
    Configuration class for Distinct L-Diversity

    :param l: Value of L to anonymize the dataset based on a column or dataset field that has a sensitive attribute. L must have a value of 2 or higher to take effect.
    :param column_name: Column or dataset field that has a sensitive attribute type.
    """

    def __init__(self, l, column_name):
        self._internal_dict = {"l": l, "column_name": column_name}
        self._anonymity_name = "LDIVERSITY_DISTINCT"
        self._print_message = f"LDiversityDistinct(l={l}, column_name={column_name})"


class LDiversityShannonEntropy(PrivacyModel):
    """
    Configuration class for Shannon Entropy L-Diversity

    :param l: Value of L to anonymize the dataset based on a column or dataset field that has a sensitive attribute. L must have a value of 2 or higher to take effect.
    :param column_name: Column or dataset field that has a sensitive attribute type.
    """

    def __init__(self, l, column_name):
        self._internal_dict = {"l": l, "column_name": column_name}
        self._anonymity_name = "LDIVERSITY_SHANNONENTROPY"
        self._print_message = f"LDiversityShannonEntropy(l={l}, column_name={column_name})"


class LDiversityGrassbergerEntropy(PrivacyModel):
    """ Configuration class for Grassberger Entropy L-Diversity

    :param l: Value of L to anonymize the dataset based on a column or dataset field that has a sensitive attribute. L must have a value of 2 or higher to take effect.
    :param column_name: Column or dataset field that has a sensitive attribute type.
    """

    def __init__(self, l, column_name):
        self._internal_dict = {"l": l, "column_name": column_name}
        self._anonymity_name = "LDIVERSITY_GRASSBERGERENTROPY"
        self._print_message = f"LDiversityGrassbergerEntropy(l={l}, column_name={column_name})"


class LDiversityRecursive(PrivacyModel):
    """ Configuration class for Recursive L-Diversity

    :param l: Value of L to anonymize the dataset based on a column or dataset field that has a sensitive attribute. L must have a value of 2 or higher to take effect.
    :param c: Value of C to anonymize the dataset based on a column or dataset field that has a sensitive attribute. c must have a value of  0.00001 or higher to take effect.
    :param column_name: Column or dataset field that has a sensitive attribute type.
    """

    def __init__(self, l, c, column_name):
        self._internal_dict = {"l": l, "c": c, "column_name": column_name}
        self._anonymity_name = "LDIVERSITY_RECURSIVE"
        self._print_message = f"LDiversityRecursive(l={l}, c={c}, column_name={column_name})"

class TClosenessOrderedDistance(PrivacyModel):
    """
    Configuration class for Ordered Distance T-Closeness

    :param t: Value of T to anonymize the dataset based on a column or dataset field that has a sensitive attribute. T must have a value between 0.000001 to 1.0
    :param column_name: Column or dataset field that has a sensitive attribute type.
    """

    def __init__(self, t, column_name):
        self._internal_dict = {"t": t, "column_name": column_name}
        self._anonymity_name = "TCLOSENESS_ORDERED_DISTANCE"
        self._print_message = f"TClosenessOrderedDistance(t={t}, column_name={column_name})"

class TClosenessEqualDistance(PrivacyModel):
    """
    Configuration class for Equal Distance T-Closeness

    :param t: Value of T to anonymize the dataset based on a column or dataset field that has a sensitive attribute. T must have a value between 0.000001 to 1.0
    :param column_name: Column or dataset field that has a sensitive attribute type.
    """

    def __init__(self, t, column_name):
        self._internal_dict = {"t": t, "column_name": column_name}
        self._anonymity_name = "TCLOSENESS_EQUAL_DISTANCE"
        self._print_message = f"TClosenessEqualDistance(t={t}, column_name={column_name})"


import copy
from collections.abc import Mapping


class AnonymizationMetrics:
    """ Understand metrics from a anonymization process"""

    def __init__(self, metrics: Mapping):
        self._elapsed_time = self._fetch_elapsed_time(metrics)
        self._attribute_generalization = self._fetch_attribute_generalization(metrics)
        self._privacy_models = self._fetch_privacy_models(metrics)

    def _fetch_elapsed_time(self, metrics):
        return metrics["processTimeMillisecounds"]

    def _fetch_attribute_generalization(self, metrics):
        return metrics["attributeGeneralization"]

    def _fetch_privacy_models(self, metrics):
        return metrics["privacyModels"]

    @property
    def attribute_generalization(self):
        """
        Property for getting the attribute generalization

        :return: Returns a dict containing all the quasi-identifying attributes and the transformation level used to anonymize the dataset.
        """
        return copy.deepcopy(self._attribute_generalization)

    @property
    def privacy_models(self):
        """
        Property for getting the privacy models with the configurations used in the anonymization process.

        :return: Returns a dict containing all the privacy models with the configuration used.
        """
        return copy.deepcopy(self._privacy_models)

    @property
    def elapsed_time(self):
        """
        Property for getting the elapsed time for the anonymization process

        :return: Returns the elapsed time in miliseconds
        """
        return self._elapsed_time

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        am_hash = hash(self.elapsed_time)
        for list_of_maps in (self._privacy_models, self._attribute_generalization):
            am_hash = hash(am_hash + self._hash_from_list_with_map(list_of_maps))
        return am_hash

    def _hash_from_list_with_map(self, list_of_items):
        am_hash = hash("")
        for mapping in list_of_items:
            am_hash = hash(am_hash + self._hash_from_mapping(mapping))
        return am_hash

    def _hash_from_mapping(self, mapping):
        am_hash = hash("")
        for metric, value in mapping.items():
            am_hash = hash(am_hash + hash(metric) + hash(str(value)))
        return am_hash






import unittest
import builtins

import pandas

from pyarxaas import state_printer


def mock_print(*args):
    pass


builtins.print = mock_print # mock print function


class StatePrinterTest(unittest.TestCase):

    def setUp(self):
        self.test_df = pandas.DataFrame()
        self.test_name_dataframe_mapping ={"zipcode": pandas.DataFrame()}

    def test_print_mapping__run(self):
        state_printer.print_mapping(self.test_name_dataframe_mapping)

    def test_jupyter_print_mapping(self):
        state_printer.jupyter_print_mapping(self.test_name_dataframe_mapping)





import unittest

import pandas

from pyarxaas.models.attribute_type import AttributeType
from pyarxaas.models.dataset import Dataset
from tests.pyarxaas import data_generator


class DatasetTest(unittest.TestCase):

    def setUp(self):
        self.test_data = [['id', 'name'],
                          ['0', 'Viktor'],
                          ['1', 'Jerry']]
        self.test_attribute_type_mapping = {'id': AttributeType.IDENTIFYING,
                                            'name': AttributeType.QUASIIDENTIFYING}

    def test_init(self):
        Dataset(self.test_data, self.test_attribute_type_mapping)

    def test_equality(self):
        dataset_1 = data_generator.id_name_dataset()
        dataset_2 = data_generator.id_name_dataset()
        self.assertEqual(dataset_1, dataset_2)
        self.assertIsNot(dataset_1, dataset_2)
        dataset_2._set_attribute_type("id", AttributeType.QUASIIDENTIFYING)
        self.assertNotEqual(dataset_1, dataset_2)

    def test_hash(self):
        dataset_1 = data_generator.id_name_dataset()
        dataset_2 = data_generator.id_name_dataset()
        test_set = {dataset_1, dataset_2}
        self.assertEqual(1, len(test_set))

    def test_init__without_attribute_types_param(self):
        dataset = Dataset(self.test_data)
        self.assertEqual(dataset._DEFAULT_ATTRIBUTE_TYPE.value, dataset._attributes[0].type.value)
        self.assertEqual(self.test_data[0][0], dataset._attributes[0].name)
        self.assertEqual(self.test_data[0][1], dataset._attributes[1].name)

    def test_create_from_pandas_dataframe(self):
        dataframe = pandas.DataFrame(self.test_data[1:], columns=self.test_data[0])
        dataset = Dataset.from_pandas(dataframe)
        pandas_df = dataset.to_dataframe()

        # assert column names are in top row
        self.assertEqual(dataframe.to_dict(), pandas_df.to_dict())

        # assert default AttributeType is set
        self.assertEqual(Dataset._DEFAULT_ATTRIBUTE_TYPE.value, dataset._attributes[0].type.value)

    def test_set_attribute_types_default_value(self):
        dataset = Dataset(self.test_data)
        self.assertEqual(AttributeType.QUASIIDENTIFYING.value, dataset._attributes[0].type.value)
        self.assertEqual(AttributeType.QUASIIDENTIFYING.value, dataset._attributes[1].type.value)

    def test_set_attribute_type_with_sequence_of_attributes(self):
        dataset = Dataset(self.test_data)
        dataset.set_attribute_type(AttributeType.IDENTIFYING, "id", "name")
        self.assertEqual(AttributeType.IDENTIFYING.value, dataset._attributes[0].type.value)
        self.assertEqual(AttributeType.IDENTIFYING.value, dataset._attributes[1].type.value)

    def test_set_attribute_type_with_single_attribute(self):
        dataset = Dataset(self.test_data)
        dataset.set_attribute_type(AttributeType.IDENTIFYING, "id")
        self.assertEqual(AttributeType.IDENTIFYING.value, dataset._attributes[0].type.value)

    def test_set_attribute_type__single_attribute(self):
        dataset = Dataset(self.test_data)
        dataset._set_attribute_type("id", AttributeType.QUASIIDENTIFYING)
        self.assertEqual(AttributeType.QUASIIDENTIFYING.value, dataset._attributes[0].type.value)
        self.assertEqual(Dataset._DEFAULT_ATTRIBUTE_TYPE.value, dataset._attributes[1].type.value)

    def test_set_hierarchy(self):

        test_hierarchy = [["0", "*"], ["1", "*"]]

        dataset = Dataset(self.test_data)
        dataset._set_attribute_type("id", AttributeType.QUASIIDENTIFYING)
        dataset.set_hierarchy("id", test_hierarchy)
        self.assertEqual(dataset._attributes[0].hierarchy, test_hierarchy)

    def test_set_hierarchy__not_valid_attribute_name(self):

        test_hierarchy = [["0", "*"], ["1", "*"]]

        dataset = Dataset(self.test_data)
        dataset._set_attribute_type("id", AttributeType.QUASIIDENTIFYING)
        with self.assertRaises(KeyError):
            dataset.set_hierarchy("fail", test_hierarchy)
        self.assertIsNone(dataset._attributes[0].hierarchy)

    def test_set_hierarchy__not_valid_attribute_type(self):

        test_hierarchy = [["0", "*"], ["1", "*"]]

        dataset = Dataset(self.test_data)
        dataset._set_attribute_type("id", AttributeType.INSENSITIVE)
        with self.assertRaises(ValueError):
            dataset.set_hierarchy("id", test_hierarchy)
        self.assertIsNone(dataset._attributes[0].hierarchy)
        self.assertIsNot(test_hierarchy, dataset._attributes[0].hierarchy)

    def test_set_hierarchies(self):

        test_hierarchy_id = [["0", "*"], ["1", "*"]]
        test_hierarchy_name = [["Viktor", "*"], ["Jerry", "*"]]

        dataset = Dataset(self.test_data)
        dataset._set_attribute_type("id", AttributeType.QUASIIDENTIFYING)
        dataset._set_attribute_type("name", AttributeType.QUASIIDENTIFYING)
        dataset.set_hierarchies({"id": test_hierarchy_id, "name": test_hierarchy_name})
        self.assertEqual(dataset._attributes[0].hierarchy, test_hierarchy_id)
        self.assertEqual(dataset._attributes[1].hierarchy, test_hierarchy_name)

    def test_set_hierarchy_with_pandas(self):

        test_hierarchy = [["0", "*"], ["1", "*"]]
        hierarchy_df = pandas.DataFrame(test_hierarchy)

        dataset = Dataset(self.test_data)
        dataset._set_attribute_type("id", AttributeType.QUASIIDENTIFYING)
        dataset.set_hierarchy("id", hierarchy_df)
        self.assertEqual(dataset._attributes[0].hierarchy, test_hierarchy)

    def test__payload(self):
        dataset = Dataset(self.test_data)
        payload = dataset._payload()
        self.assertEqual(AttributeType.QUASIIDENTIFYING.value, payload["attributes"][0]["attributeTypeModel"])
        self.assertEqual(None, payload["attributes"][0]["hierarchy"])

    def test__payload__with_hierarchies(self):

        test_hierarchy_id = [["0", "*"], ["1", "*"]]
        test_hierarchy_name = [["Viktor", "NAME"], ["Jerry", "NAME"]]

        dataset = Dataset(self.test_data)
        dataset._set_attribute_type("id", AttributeType.QUASIIDENTIFYING)
        dataset._set_attribute_type("name", AttributeType.QUASIIDENTIFYING)
        dataset.set_hierarchies({"id": test_hierarchy_id, "name": test_hierarchy_name})
        payload = dataset._payload()
        self.assertEqual(test_hierarchy_id, payload["attributes"][0]["hierarchy"])
        self.assertEqual(test_hierarchy_name, payload["attributes"][1]["hierarchy"])

    def test_to_dataframe(self):
        dataset = Dataset(self.test_data, self.test_attribute_type_mapping)
        df = dataset.to_dataframe()
        self.assertIsInstance(df, pandas.DataFrame)

    def test_from_dict(self):
        data = {"id": [1, 2], "name": ["Monsen", "Mikkel"]}
        expected_df = pandas.DataFrame.from_dict(data)
        dataset = Dataset.from_dict(data)
        self.assertIsNotNone(dataset)
        self.assertIsInstance(dataset, Dataset)
        self.assertEqual(expected_df.to_dict(), dataset.to_dataframe().to_dict())





import copy
from collections import Mapping

from pandas import DataFrame


class RiskProfile:
    """
    Represents the re-identification risks associated with a Dataset
    """

    def __init__(self, metrics: Mapping):
        self._re_identification_of_risk = copy.deepcopy(metrics["reIdentificationRisk"]["measures"])
        self._distribution_of_risk = copy.deepcopy(metrics["distributionOfRisk"])
        self._attacker_success_rate = copy.deepcopy(metrics["reIdentificationRisk"]["attackerSuccessRate"]["successRates"])
        self._quasi_identifiers = metrics["reIdentificationRisk"]["quasiIdentifiers"]
        self._population_model = metrics["reIdentificationRisk"]["populationModel"]

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(self._metric_hash(self._re_identification_of_risk)
                    + self._metric_hash(self._distribution_of_risk["riskIntervalList"][0]))

    def _metric_hash(self, metric):
        m_hash = hash("")
        for metric, value in metric.items():
            m_hash = hash(m_hash + hash(hash(metric) + hash(str(value))))
        return m_hash


    @property
    def re_identification_risk(self):
        """
        Re-identification risk metrics for a given Dataset

        :return: dict containing re-identification metrics
        """
        return copy.deepcopy(self._re_identification_of_risk)

    @property
    def distribution_of_risk(self):
        """
        Distribution of risk for a given Dataset

        :return: dict containing the distribution of risks in a given Dataset
        """
        return copy.deepcopy(self._distribution_of_risk)

    @property
    def attacker_success_rate(self):
        """
        Attacker success rates against re-identification for a given Dataset

        :return: dict containing the attacker success rate.
        """
        return copy.deepcopy(self._attacker_success_rate)

    @property
    def quasi_identifiers(self):
        """
        Quasi-identifiers for a given Dataset

        :return: dict containing a list of all the quasi-identifying attribute in a a given Dataset
        """
        return copy.deepcopy(self._quasi_identifiers)

    @property
    def population_model(self):
        """
        Population model used to analyze a given Dataset

        :return: The Population model name used to analyze a given Dataset
        """
        return self._population_model

    def re_identification_risk_dataframe(self) -> DataFrame:
        """
        Re-identification risk as a pandas.DataFrame

        :return: pandas.Dataframe with risk metrics
        """
        df = DataFrame([self._re_identification_of_risk])
        return df

    def distribution_of_risk_dataframe(self) -> DataFrame:
        """
        Distribution of risk as a pandas.DataFrame

        :return: pandas.DataFrame
        """
        return DataFrame.from_dict(self._distribution_of_risk["riskIntervalList"])



import uuid


class Level:
    """
    Represent a given level in a hierarchy builder
    """

    def __init__(self, level):
        self._level = level
        self._groups = {}

    def add_group(self, grouping, label=None) -> "Level":
        self._groups[Group(grouping, label)] = None
        return self

    def payload(self):
        return {"level": self._level,
                "groups": [group.payload() for group in self._groups.keys()]}

    def __repr__(self):
        return f"Level(level={self._level}, groups={self._groups})"

    def __str__(self):
        return f"level={self._level}, groups={self._groups}"


class Group:
    """
    Represents a group in a hierarchy Level
    """

    def __init__(self, grouping, label):
        self._grouping = grouping
        self._label = label

    def payload(self):
        return {"grouping": self._grouping,
                "label": self._label}

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        # Groups are only equal if labels are present and the same
        label = self._label
        if label is None:
            label = uuid.uuid1()
        return hash(self._grouping) + hash(label)

    def __repr__(self):
        return f"Group(grouping={self._grouping}, label={self._label})"

    def __str__(self):
        return f"grouping={self._grouping}, label={self._label}"








import copy
from abc import ABC

from pyarxaas.hierarchy.level import Level


class GroupingBasedHierarchy(ABC):

    def __init__(self):
        self._levels = []
        self._level_count = 0

    @property
    def levels(self):
        return copy.deepcopy(self._levels)

    def level(self, level) -> Level:
        """
        Get a level in the hierarchy

        :param level: int representing the level
        :return: Level
        """
        if level > self._level_count:
            raise AttributeError(f"level={level} is higher than available levels={self._level_count}")
        if level == len(self._levels):
            self._level_count += 1
            self._levels.append(Level(level))
        return self._levels[level]


import unittest
from pyarxaas.models.anonymization_metrics import AnonymizationMetrics
from tests.pyarxaas import data_generator


class AnonymizationMetricsTest(unittest.TestCase):

    def setUp(self):
        self.test_metrics = data_generator.raw_anonymization_metrics()

    def test_init(self):
       self.assertIsNotNone(AnonymizationMetrics(self.test_metrics))

    def test_elapsed_time(self):
        am = AnonymizationMetrics(self.test_metrics)
        actual = self.test_metrics["processTimeMillisecounds"]
        self.assertEqual(actual, am.elapsed_time)

    def test_privacy_models(self):
        am = AnonymizationMetrics(self.test_metrics)
        self.assertEqual(self.test_metrics["privacyModels"], am.privacy_models)
        self.assertIsNot(self.test_metrics["privacyModels"], am.privacy_models)

    def test_attribute_generalization(self):
        am = AnonymizationMetrics(self.test_metrics)
        self.assertEqual(self.test_metrics["attributeGeneralization"], am.attribute_generalization)
        self.assertIsNot(self.test_metrics["attributeGeneralization"], am.attribute_generalization)

    def test_equality(self):
        am1 = AnonymizationMetrics(self.test_metrics)
        am2 = AnonymizationMetrics(self.test_metrics)
        self.assertEqual(am1, am2)
        self.assertIsNot(am1, am2)
        am2._elapsed_time = 200
        self.assertNotEqual(am1, am2)

    def test_hashcode(self):
        am1 = AnonymizationMetrics(self.test_metrics)
        am2 = AnonymizationMetrics(self.test_metrics)
        test_set = {am1, am2}
        self.assertEqual(1, len(test_set))



import json
import unittest
from pathlib import Path

from numpy import dtype
from pandas import DataFrame

from pyarxaas.models.risk_profile import RiskProfile


class RiskProfileTest(unittest.TestCase):

    def setUp(self):
        response_path = Path(__file__).parent.joinpath("..").joinpath("test_data").joinpath("analyze_response_test_data.json")
        with response_path.open(encoding="utf-8") as file:
            json_str = file.read()
        self.risk_profile_response = json.loads(json_str)

    def test_init(self):
        RiskProfile(self.risk_profile_response)

    def test_equality(self):
        risk_profile_1 = RiskProfile(self.risk_profile_response)
        risk_profile_2 = RiskProfile(self.risk_profile_response)
        self.assertEqual(risk_profile_1, risk_profile_2)
        risk_profile_2._re_identification_of_risk["estimated_prosecutor_risk"] = 50.0
        self.assertNotEqual(risk_profile_1, risk_profile_2)

    def test_hash(self):
        risk_profile_1 = RiskProfile(self.risk_profile_response)
        risk_profile_2 = RiskProfile(self.risk_profile_response)
        test_set = {risk_profile_1, risk_profile_2}
        self.assertEqual(1, len(test_set))

    def test_to_dataframe(self):
        risk_profile = RiskProfile(self.risk_profile_response)
        df = risk_profile.re_identification_risk_dataframe()
        self.assertIsInstance(df, DataFrame)

    def test_re_identification_risk_to_dataframe_shape(self):
        risk_profile = RiskProfile(self.risk_profile_response)
        df = risk_profile.re_identification_risk_dataframe()
        self.assertEqual(self.risk_profile_response["reIdentificationRisk"]["measures"]["records_affected_by_highest_prosecutor_risk"], df["records_affected_by_highest_prosecutor_risk"][0])

    def test_distribution_of_risk_to_dataframe_shape(self):
        risk_profile = RiskProfile(self.risk_profile_response)
        df = risk_profile.distribution_of_risk_dataframe()
        self.assertEqual(self.risk_profile_response["distributionOfRisk"]["riskIntervalList"][0]["interval"],
                         df["interval"][0])

    def test_re_identification_risk_to_dataframe__column_types(self):
        risk_profile = RiskProfile(self.risk_profile_response)
        df = risk_profile.re_identification_risk_dataframe()
        for d_type in df.dtypes.tolist():
            self.assertEqual(d_type, dtype("float64"))

    def test_attacker_success_rate_property(self):
        expected = {'Prosecutor_attacker_success_rate': 1.0, 'Marketer_attacker_success_rate': 1.0, 'Journalist_attacker_success_rate': 1.0}
        risk_profile = RiskProfile(self.risk_profile_response)
        self.assertEqual(expected, risk_profile.attacker_success_rate)

    def test_quasi_indentifers_property(self):
        expected = ['zipcode']
        risk_profile = RiskProfile(self.risk_profile_response)
        self.assertEqual(expected, risk_profile.quasi_identifiers)

    def test_population_model_property(self):
        expected = "ZAYATZ"
        risk_profile = RiskProfile(self.risk_profile_response)
        self.assertEqual(expected, risk_profile.population_model)




class Interval:

    def __init__(self, from_, to, label: str = None):
        if from_ > to:
            raise AttributeError(f"from={from_} cannot be bigger than to={to}")
        self._from = from_
        self._to = to
        self._label = label

    def payload(self):
        return {"from": self._from,
                "to": self._to,
                "label": self._label}

    def is_decimal(self):
        if isinstance(self._from, float) or isinstance(self._to, float):
            return True
        return False

    def __repr__(self):
        return f"Interval(from={self._from}, to={self._to}, label={self._label})"

    def __str__(self):
        return f"from={self._from}, to={self._to}, label={self._label}"

    def __eq__(self, other):
        if isinstance(other, Interval):
            return hash(other) == hash(self)
        return False

    def __hash__(self):
        return hash(self._from) + hash(self._to) + hash(self._label)



from pyarxaas.models.dataset import Dataset
from pyarxaas.models.attribute_type import AttributeType
from pyarxaas.arxaas import ARXaaS
import logging

# Set null logger to provide override for package user
logging.getLogger(__name__).addHandler(logging.NullHandler())

# Make important classes an packages available from top level of package
__all__ = ['ARXaaS', 'Dataset', 'AttributeType']


from pyarxaas.hierarchy.interval_builder.interval_hierarchy_builder import IntervalHierarchyBuilder
from pyarxaas.hierarchy.redaction_hierarchy_builder import RedactionHierarchyBuilder
from pyarxaas.hierarchy.order_builder.order_hierarchy_builder import OrderHierarchyBuilder

__all__ = ['IntervalHierarchyBuilder', 'RedactionHierarchyBuilder', 'OrderHierarchyBuilder']


import unittest
from pyarxaas.models.attribute_type import AttributeType
from pyarxaas.models.dataset import Dataset
from pyarxaas.models.risk_profile import RiskProfile
from pyarxaas.models.anonymize_result import AnonymizeResult

class AnonymizationResultTest(unittest.TestCase):

    def setUp(self):
        self.test_data = [['id', 'name'],
                          ['0', 'Viktor'],
                          ['1', 'Jerry']]
        self.test_attribute_type_mapping = {'id': AttributeType.IDENTIFYING,
                                            'name': AttributeType.QUASIIDENTIFYING}
        self.test_dataset = Dataset(self.test_data, self.test_attribute_type_mapping)
        self.risk_profile_response = {"reIdentificationRisk": {"measures": {
            "measure_value": "[%]",
            "Prosecutor_attacker_success_rate": "98.72",
            "records_affected_by_highest_prosecutor_risk": "97.46000000000001",
            "sample_uniques": "97.46000000000001",
            "estimated_prosecutor_risk": "100.0",
            "population_model": "PITMAN",
            "highest_journalist_risk": "100.0",
            "records_affected_by_lowest_risk": "0.06",
            "estimated_marketer_risk": "98.72000000000001",
            "Journalist_attacker_success_rate": "98.72",
            "highest_prosecutor_risk": "100.0",
            "estimated_journalist_risk": "100.0",
            "lowest_risk": "33.33333333333333",
            "Marketer_attacker_success_rate": "98.72",
            "average_prosecutor_risk": "98.72000000000001",
            "records_affected_by_highest_journalist_risk": "97.46000000000001",
            "population_uniques": "39.64593493418713",
            "quasi_identifiers": ["Innvandrerbakgrunn", "Ytelse", "Innsatsgruppe", "Ledighetsstatus"]
        }},
            "distributionOfRisk": {"riskIntervalList": [{"interval": "]50,100]",
                                                         "recordsWithRiskWithinInteval": 0.9746,
                                                         "recordsWithMaxmalRiskWithinInterval": 1.0},
                                                        {"interval": "]33.4,50]",
                                                         "recordsWithRiskWithinInteval": 0.0248,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0254},
                                                        {"interval": "]25,33.4]",
                                                         "recordsWithRiskWithinInteval": 0.0006,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0006},
                                                        {"interval": "]20,25]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]16.7,20]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]14.3,16.7]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]12.5,14.3]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]10,12.5]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]9,10]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]8,9]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]7,8]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]6,7]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]5,6]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]4,5]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]3,4]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]2,3]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]1,2]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]0.1,1]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]0.01,0.1]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]0.001,0.01]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]0.0001,0.001]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]1e-5,0.0001]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]1e-6,1e-5]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0},
                                                        {"interval": "]0,1e-6]",
                                                         "recordsWithRiskWithinInteval": 0.0,
                                                         "recordsWithMaxmalRiskWithinInterval": 0.0}]
                                   }
        }



        self.test_riskprofile = RiskProfile(self.risk_profile_response)

        self.test_anonymize_result = AnonymizeResult(self.test_data, self.test_riskprofile, self.test_metrics)




import unittest

from pyarxaas.models.dataset.attribute import Attribute
from pyarxaas.models.attribute_type import AttributeType


class AttributeTest(unittest.TestCase):

    def test_name(self):
        field = Attribute("id", AttributeType.QUASIIDENTIFYING)
        self.assertEqual("id", field.name)

    def test_type(self):
        field = Attribute("id", AttributeType.QUASIIDENTIFYING)
        self.assertEqual(AttributeType.QUASIIDENTIFYING.value, field.type.value)




import unittest

from pyarxaas.hierarchy.interval_builder.interval import Interval


class IntervalTest(unittest.TestCase):

    def test_init(self):
        self.assertIsNotNone(Interval(0, 10))

    def test_equality(self):
        i1 = Interval(0, 10, "child")
        i2 = Interval(0, 10, "child")
        self.assertEqual(i1, i2)

    def test_hash(self):
        i1 = Interval(0, 10, "child")
        i2 = Interval(0, 10, "child")
        test_set = {i1, i2}
        self.assertEqual(1, len(test_set))




import unittest

import pandas

from pyarxaas.models.dataset.data import Data
from pyarxaas.models.attribute_type import AttributeType
from pyarxaas.models.dataset import Dataset
from tests.pyarxaas import data_generator


class DataTest(unittest.TestCase):

    def setUp(self):
        pass

    def test_init(self):
        Data(headers=["id", "name"], rows=[["1", "Max"]])

    def test_payload(self):
        data = Data(headers=["id", "name"], rows=[["1", "Max"]])
        self.assertEqual([["id", "name"], ["1", "Max"]], data.payload)


import unittest

import pandas

from pyarxaas import converters
from pyarxaas.privacy_models import KAnonymity, LDiversityDistinct, LDiversityRecursive


class ConvertersTest(unittest.TestCase):

    def setUp(self):
        self.test_hierarchy_mapping = {'zipcode': [['81667', '8166*', '816**', '81***', '8****', '*****'],
                                            ['81668', '8166*', '816**', '81***', '8****', '*****'],
                                            ['81669', '8166*', '816**', '81***', '8****', '*****'],
                                            ['81670', '8167*', '816**', '81***', '8****', '*****'],
                                            ['81671', '8167*', '816**', '81***', '8****', '*****'],
                                            ['81672', '8167*', '816**', '81***', '8****', '*****'],
                                            ['81673', '8167*', '816**', '81***', '8****', '*****'],
                                            ['81674', '8167*', '816**', '81***', '8****', '*****'],
                                            ['81675', '8167*', '816**', '81***', '8****', '*****'],
                                            ['81676', '8167*', '816**', '81***', '8****', '*****'],
                                            ['81677', '8167*', '816**', '81***', '8****', '*****']]}

        self.test_privacy_models_sequence = [KAnonymity(k=4), LDiversityDistinct(2, "test_column")]
        self.test_attribute_types_mapping = {"id": "INDENTIFING",
                                             "name": "QUASIIDENTIFING"}

    def test_create_transform_models_dataframe__run(self):
        dataframe = converters.create_transform_models_dataframe(self.test_hierarchy_mapping)
        self.assertIsNotNone(dataframe)
        self.assertIsInstance(dataframe, pandas.DataFrame)


    def test_create_privacy_models_dataframe__run(self):
        dataframe = converters.create_privacy_models_dataframe(self.test_privacy_models_sequence)
        self.assertIsNotNone(dataframe)
        self.assertIsInstance(dataframe, pandas.DataFrame)

    def test_create_attribute_types_dataframe__run(self):
        dataframe = converters.create_attribute_types_dataframe(self.test_attribute_types_mapping)
        self.assertIsNotNone(dataframe)
        self.assertIsInstance(dataframe, pandas.DataFrame)



from collections.abc import Sequence, Mapping

import pandas

from pyarxaas.privacy_models import PrivacyModel


def create_privacy_models_dataframe(privacy_models: Sequence) -> pandas.DataFrame:
    """
    Creates a pandas.DataFrame from Sequence of PrivacyModels objects

    :param privacy_models: Sequence of PrivacyModels
    :return: pandas.DataFrame
    """
    privacy_models_index = []
    privacy_models_values = []

    for model in privacy_models:
        model_index = create_model_indexes(model)
        model_values = create_model_values(model)
        privacy_models_index += model_index
        privacy_models_values += model_values
    index = pandas.MultiIndex.from_tuples(privacy_models_index, names=("privacy_model", "parameter"))
    dataframe = pandas.DataFrame(privacy_models_values, index=index, columns=("value",))
    return dataframe


def create_model_indexes(model: PrivacyModel):
    """
    Creates the indexes for a provided PrivacyModel to be used in a MultiIndex DataFrame
    :param model: PrivacyModel object
    :return: model_index: tuple of tuples containing indexes
    """
    model_index = []
    for key, value in model.items():
        model_index.append((model.name, key))
    return model_index


def create_model_values(model: PrivacyModel):
    """
    Create list of values from the PrivacyModel object
    :param model: PrivacyModel object
    :return: List of values
    """
    model_values = []
    for value in model.values():
        model_values.append(value)
    return model_values


def create_attribute_types_dataframe(attribute_types: Mapping) -> pandas.DataFrame:
    """
    Creates a pandas.Dataframe from a Mapping of attribute_types str:field=str:attribute_type

    :param attribute_types: Mapping str:field=str:attribute_type
    :return: pandas.DataFrame
    """
    attribute_rows = attribute_types.items()
    return pandas.DataFrame(attribute_rows, columns=("field", "type")).set_index("field")


def create_transform_models_dataframe(transform_models):
    """
    Creates a pandas.DataFrame form a Mapping of Transform Models


    :param transform_models: mapping
    :return: pandas.DataFrame
    """
    transform_models_index = []
    transform_models_values = []

    for field, hierarchy in transform_models.items():
        field_index, field_values = _create_index_and_values_for(field, hierarchy)
        transform_models_index += field_index
        transform_models_values += field_values

    index = pandas.MultiIndex.from_tuples(transform_models_index)
    dataframe = pandas.DataFrame(transform_models_values, index=index)
    return dataframe


def _create_index_and_values_for(field, hierarchy):
    transform_model_index = []
    transform_model_values = []

    for level in range(1, _get_hierarchy_levels(hierarchy)):
        model_index = [field]
        level_values = []
        model_index.append(f"level_{level}")
        transform_model_index.append(model_index)
        for row in hierarchy:
            level_values.append(row[level])
        transform_model_values.append(level_values)
    return transform_model_index, transform_model_values


def create_dataframe_with_index_from_mapping(mapping: Mapping, columns: Sequence):
    """
    Convenicence function for easy creation of pandas.DataFrame from Python mapping object
    :param mapping: Mapping object to create pandas.DataFrame from
    :param columns: Columns to use in the DataFrame. First column will be used as index
    :return: pandas.DataFrame
    """
    return pandas.DataFrame(mapping.items(), columns=columns).set_index(columns[0])


def _get_hierarchy_levels(hierarchy):
    return len(hierarchy[0])



from collections.abc import Mapping
from enum import Enum


class RedactionHierarchyBuilder:
    """
    Understands building redaction based hierarchies
    """

    class Order(Enum):
        LEFT_TO_RIGHT = "LEFT_TO_RIGHT"
        RIGHT_TO_LEFT = "RIGHT_TO_LEFT"

    def __init__(self, padding_char:str = " ",
                 redaction_char: str = "*",
                 padding_order: Order = Order.RIGHT_TO_LEFT,
                 redaction_order: Order = Order.RIGHT_TO_LEFT):
        self._assert_padding_is_valid(redaction_char, padding_char)
        self._padding_char = padding_char
        self._reduction_char = redaction_char
        self._padding_order = padding_order
        self._redaction_order = redaction_order

    def _request_payload(self)-> Mapping:
        return {
            "builder": {
                "type": "redactionBased",
                "paddingCharacter": self._padding_char,
                "redactionCharacter": self._reduction_char,
                "paddingOrder": self._padding_order.value,
                "redactionOrder": self._redaction_order.value
            }
        }

    def _assert_padding_is_valid(self, *characters):
        for char in characters:
            if len(char) > 1:
                raise AttributeError(f"characters can only be a single character: len({char}) > 1")



import copy
from collections.abc import Sequence
from collections.abc import Mapping

import pandas

from pyarxaas.models.dataset.data import Data
from pyarxaas.models.dataset.attribute import Attribute
from pyarxaas.models.attribute_type import AttributeType


class Dataset:
    """
    Understand tabular data containing personal data.
    """

    _DEFAULT_ATTRIBUTE_TYPE = AttributeType.QUASIIDENTIFYING

    def __init__(self, data: list, attribute_types: Mapping = None):
        if attribute_types is None:
            attribute_types = self._create_default_attribute_map(data[0])

        self._data = Data(data[0], data[1:])
        self._attributes = self._create_attributes(attribute_types)

    def _set_attribute_type(self, attribute, attribute_type: AttributeType):
        """
        Set Attribute type for a attribute in the dataset

        :param attribute: attribute in the dataset
        :param attribute_type: AttributeType for the attribute
        :return: None
        """

        field_map = {field.name: field for field in self._attributes}
        try:
            field_map[attribute].type = attribute_type
        except KeyError:
            raise KeyError(f"attribute=({attribute}) could not be found")

    def set_attribute_type(self, attribute_type: AttributeType, *attributes):
        """
        Set AttributeType for a collection of attributes

        :param attributes: collection of attributes in the dataset
        :param attribute_type: AttributeType for the attributes
        :return: None
        """
        for attribute in attributes:
            self._set_attribute_type(attribute, attribute_type)

    def _create_attributes(self, attribute_types: Mapping):
        fields = []
        for field_name, type in attribute_types.items():
            fields.append(Attribute(field_name, type))
        return fields

    def set_hierarchy(self, attribute, hierarchy):
        """
        Set hierarchy for a attribute in the Dataset

        :param attribute: attribute in the Dataset
        :param hierarchy: to be applied  to the attribute
        :return: None
        """

        hierarchy = self._create_from_hierarchy_source(hierarchy)
        field_map = {field.name: field for field in self._attributes}
        try:
            field_map[attribute].hierarchy = hierarchy
        except KeyError:
            raise KeyError(f"attribute=({attribute}) could not be found")

    def set_hierarchies(self, hierarchies):
        for attribute, hierarchy in hierarchies.items():
            self.set_hierarchy(attribute, hierarchy)

    def to_dataframe(self) -> pandas.DataFrame:
        """
        Create pandas DataFrame of the Dataset

        :return: pandas.DataFrame
        """

        return self._data.dataframe

    def describe(self):
        """
        Prints a description of the Dataset to stdout

        :return: None
        """

        indent = 2
        self._data.describe(indent)
        print("attributes:")
        print(self._describe_attributes(indent))

    def _describe_attributes(self, indent):
        string = ""
        for attribute in self._attributes:
            string += " "*indent + str(attribute) + "\n"
        return string

    def _payload(self):
        payload = {}
        dataset_dict = self._to_dict()
        payload["data"] = dataset_dict["data"]
        payload["attributes"] = dataset_dict["attributes"]
        return payload

    def _to_dict(self):
        return {
            "data": self._data.payload,
            "attributes": self._create_attributes_payload()
        }

    def _create_attributes_payload(self):
        attributes = []
        for field in self._attributes:
            attributes.append(field.payload)
        return attributes

    @classmethod
    def from_pandas(cls, dataframe: pandas.DataFrame):
        """
        Create a Dataset from a pandas DataFrame

        :param dataframe: pandas Dataframe
        :return: Dataset
        """

        headers = dataframe.columns.values.tolist()
        values = dataframe.values.tolist()
        data = [headers] + values
        return Dataset(data=data, attribute_types=cls._create_default_attribute_map(headers))

    @classmethod
    def from_dict(cls, dictionary):
        """
        Create Dataset from a python dictionary

        :param dictionary: Mapping object to create Dataset from
        :return: Dataset
        """

        df = pandas.DataFrame.from_dict(dictionary)
        return cls.from_pandas(df)

    @classmethod
    def _create_default_attribute_map(cls, fields):
        attribute_type_map = {}
        for field in fields:
            attribute_type_map[field] = cls._DEFAULT_ATTRIBUTE_TYPE
        return attribute_type_map

    @staticmethod
    def _create_from_hierarchy_source(source):
        if isinstance(source, Sequence):
            return copy.deepcopy(source)
        if isinstance(source, pandas.DataFrame):
            return source.values.tolist()

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(hash(self._data) + self._hash_of_attributes())

    def _hash_of_attributes(self):
        a_hash = hash(self._attributes[0])
        for attribute in self._attributes[0:]:
            a_hash = hash(a_hash + hash(attribute))
        return a_hash

    def __repr__(self) -> str:
        return f"Dataset(data={self._data}, attributes={self._attributes})"

















"""
A setuptools based setup module.
"""


from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'VERSION'), encoding='utf-8') as f:
    version = f.read()

setup(

    name='pyARXaaS',  # Required

    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='ARXaaS Python wrapper package',  # Optional

    long_description=long_description,  # Optional

    long_description_content_type='text/markdown',  # Optional

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/oslomet-arx-as-a-service/PyARXaaS',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='OsloMET-gruppe-8',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='sondre.hal@gmail.com',  # Optional

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='anonymization ARX ARXaaS ARX-as-a-Service',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["uplink==0.7.0",
                      "pandas==0.24.2",
                      "IPython==7.5.0"],  # Optional

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={  # Optional
        'test': ['coverage'],
    },


    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/oslomet-arx-as-a-service/PyARXaaS/issues',
        'Source': 'https://github.com/oslomet-arx-as-a-service/PyARXaaS',
    }
)



from pyarxaas.hierarchy.grouping_based_hierarchy import GroupingBasedHierarchy


class OrderHierarchyBuilder(GroupingBasedHierarchy):

    def __init__(self):
        super().__init__()

    def _request_payload(self):
        return {
            "builder": {
                "type": "orderBased",
                "levels": [level.payload() for level in self.levels]
            }
        }



import copy

from pyarxaas.models.anonymization_metrics import AnonymizationMetrics
from pyarxaas.models.dataset import Dataset
from pyarxaas.models.risk_profile import RiskProfile


class AnonymizeResult:
    """ Understands the result of a anonymization process"""

    def __init__(self, dataset: Dataset, risk_profile: RiskProfile, anonymization_metrics: AnonymizationMetrics, anonymization_status):
        self._anonymization_metrics = anonymization_metrics
        self._risk_profile = risk_profile
        self._dataset = dataset
        self._anonymization_status = anonymization_status

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(hash(self._anonymization_metrics) + hash(self._risk_profile) + hash(self._dataset))

    @property
    def dataset(self) -> Dataset:
        """
        Dataset created from the anonymization

        :return: Dataset
        """
        return copy.deepcopy(self._dataset)

    @property
    def risk_profile(self) -> RiskProfile:
        """
        RiskProfile asscocciated with the new Dataset

        :return: RiskProfile
        """
        return copy.deepcopy(self._risk_profile)

    @property
    def anonymization_metrics(self) -> AnonymizationMetrics:
        """
        AnonymizationMetrics about the anonymization process.
        Contains data on hierarchy level used and privacy model configuration

        :return: AnonymizationMetrics
        """
        return copy.deepcopy(self._anonymization_metrics)

    @property
    def anonymization_status(self) -> str:
        """
        Anonymization status for the new Dataset

        :return: str
        """
        return self._anonymization_status

    @classmethod
    def _from_response(cls, dataset, risk_profile, anon_metrics, anon_status):
        anonymize_metrics = AnonymizationMetrics(anon_metrics)
        return cls(dataset, risk_profile, anonymize_metrics, anon_status)



